/*******************************************************************
                      C O P Y R I G H T  (c) 1998
        F O C U S     A U T O M A T I O N   S Y S T E M S   I N C.
                          All Rights Reserved.
 -------------------------------------------------------------------

 verify.c - A set of modify-verify callbacks for Photon widgets.

 Version 1.1; - added ufloat and fufloat callbacks.
              - fixed buggy handling of negative numbers that were
                pre-entered in unsigned fields.

 -------------------------------------------------------------------
 *******************************************************************
 * $Source: /usr/local/cvsroot/fasi/jobs/j240/src/con_mgr/src/verify.c,v $
 * $Date: 1998/03/03 01:38:36 $
 * $Revision: 1.1 $
 *******************************************************************
*/

/* Standard headers */
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <ctype.h>
#include <math.h>

/* Toolkit headers */
#include <Ph.h>
#include <Pt.h>
#include <Ap.h>

/* Local headers */
#include "abimport.h"
#include "proto.h"

/* If ELECTROSTATIC is defined, the verify callbacks will automatically
   attach themselves all over the widget when they are attached as
   Pt_CB_REALIZED callbacks.  This is not very clean software architecture,
   but since the verify callbacks are useful for 4 different callback
   types, and since you may have dialogs full of text entry fields to
   verify, it is just much easier to only manually attach the callbacks
   once in PhAB and then let the code do the work. If you don't like the
   behaviour, turn off the define.  Easy. */

#if 1
    #define ELECTROSTATIC
#endif


#ifdef ELECTROSTATIC

/* This little function takes a PtLinkedList_t list and packs its contents
   into an array. The list is unchanged, and the array is unbounded.       */

static int get_list( const PtLinkedList_t *list, void *array, size_t size )
{
    int count;
    char *char_array;

    count = 0;

    char_array = array;

    for ( ; list; list = list->next )
    {
        memcpy( (char_array + size * count), &list->data, size );

        count++;
    }

    return ( count );
}


static void attach_callbacks( PtWidget_t *widget, void *func )
{
    PtArg_t         args[20];
    int             nargs;
    int             attached;
    PtLinkedList_t *list;
    PtLinkedList_t *verify_list;
    PtLinkedList_t *focus_list;
    PtLinkedList_t *activate_list;
    PtCallback_t   *callbackp;
    PtCallback_t    verify_array[40];
    PtCallback_t    focus_array[40];
    PtCallback_t    activate_array[40];
    long            verify_count;
    long            focus_count;
    long            activate_count;

    /* First, check to see if the callbacks have been attached during
       a previous realize operation. */

    attached = Pt_FALSE;

    PtSetArg( &args[0], Pt_CB_MODIFY_VERIFY, &list, 0 );

    PtGetResources( widget, 1, &args );

    for ( ; list != NULL; list = list->next )
    {
        callbackp = (PtCallback_t *)&list->data;

        if ( callbackp->event_f == func )
        {
            attached = Pt_TRUE;
            break;
        }
    }

    /* Now the callbacks can be attached. */

    if ( !attached )
    {
        nargs = 0;

        PtSetArg( &args[nargs++], Pt_CB_MODIFY_VERIFY, &verify_list,   0 );
        PtSetArg( &args[nargs++], Pt_CB_LOST_FOCUS,    &focus_list,    0 );
        PtSetArg( &args[nargs++], Pt_CB_ACTIVATE,      &activate_list, 0 );

        PtGetResources( widget, nargs, &args );

        nargs = 0;

        verify_array[0].event_f   = func;
        verify_array[0].data      = NULL;
        focus_array[0].event_f    = func;
        focus_array[0].data       = NULL;
        activate_array[0].event_f = func;
        activate_array[0].data    = NULL;

        verify_count   = 1 + get_list( verify_list,   &verify_array[1],
                                       sizeof(verify_array[0]) );
        focus_count    = 1 + get_list( focus_list,    &focus_array[1],
                                       sizeof(focus_array[0]) );
        activate_count = 1 + get_list( activate_list, &activate_array[1],
                                       sizeof(activate_array[0]) );

        PtSetArg( &args[nargs++], Pt_CB_MODIFY_VERIFY, &verify_array,
                                                        verify_count );
        PtSetArg( &args[nargs++], Pt_CB_LOST_FOCUS, &focus_array,
                                                     focus_count );
        PtSetArg( &args[nargs++], Pt_CB_ACTIVATE, &activate_array,
                                                   activate_count );

        PtSetResources( widget, nargs, &args );
    }

    return;
}
#endif /* ELECTROSTATIC */


/* make_str() builds the string that will be created if a modify-verify
   callback approves the current text change.  Verifying that this string
   is correct is often easier than working with the nasty data structures
   that Photon gives you.  See the "Callback structure" section of the
   PtText online help for a description of the data structures this
   function works with. */

static void make_str( PtWidget_t *widget, PtTextCallback_t *data, char *buff )
{
    PtArg_t  arg;
    char    *text;
    char     tempbuff[120];
    int      cur_insert;

    *buff = NULL;

    PtSetArg( &arg, Pt_ARG_TEXT_STRING, &text, 0 );
    PtGetResources( widget, 1, &arg );


    /* Copy the current text into a temporary buffer,
       performing any text deletion that is required. */

    if ( data->start_pos != data->end_pos )
    {
        if ( data->start_pos > 0 )
        {
            memcpy( tempbuff, text, data->start_pos );
            tempbuff[ data->start_pos ] = NULL;
        }
        else
        {
            tempbuff[0] = NULL;
        }

        strcat( tempbuff, &text[ data->end_pos ] );
    }
    else
    {
        if ( text )
        {
            strcpy( tempbuff, text );
        }
        else
        {
            tempbuff[0] = NULL;
        }
    }

    /* Now copy the temporary buffer contents into the output
       string, performing any text additions that are required. */

    if ( data->length > 0 )
    {
        cur_insert = min( data->cur_insert, strlen(tempbuff) );

        if ( cur_insert > 0 )
        {
            memcpy( buff, tempbuff, cur_insert );
        }

        memcpy( &buff[ cur_insert ], data->text, data->length );
        buff[ cur_insert + data->length ] = NULL;

        strcat( buff, &tempbuff[ cur_insert ] );
    }
    else
    {
        strcpy( buff, tempbuff );
    }

    return;
}


/* get_text() extracts the text string from a widget. This function is
   only used because the way of getting the text string from a widget
   depends on what kind of callback is used. Supported callback types
   are Pt_CB_LOST_FOCUS, Pt_CB_ACTIVATE and Pt_CB_REALIZED. */

static char *get_text( PtWidget_t *widget, PtCallbackInfo_t *cbinfo )
{
    char    *text = NULL;
    PtArg_t  arg;

    if ( cbinfo->reason == Pt_CB_LOST_FOCUS )
    {
        text = ((PtTextCallback_t *)cbinfo->cbdata)->text;
    }
    else
    if ( cbinfo->reason == Pt_CB_ACTIVATE &&
         cbinfo->reason_subtype == Pt_EDIT_ACTIVATE )
    {
        text = ((PtTextCallback_t *)cbinfo->cbdata)->text;
    }
    else
    if ( cbinfo->reason == Pt_CB_ACTIVATE ||
         cbinfo->reason == Pt_CB_REALIZED    )
    {
        PtSetArg( &arg, Pt_ARG_TEXT_STRING, &text, 0 );
        PtGetResources( widget, 1, &arg );
    }

    return ( text );
}


/*
-----------------------------------------------------------------------
int verify_int( widget, apinfo, cbinfo ) - verify integer values.

As a MODIFY_VERIFY callback:
Verifies that a text field contains a valid integer value.  An integer
can start with a '-' character, but is otherwise all digits.

As a LOST_FOCUS or ACTIVATE callback:
Reformats the integer (if necessary) to look like a normal integer
value. '-000' becomes '0', for example. LOST_FOCUS applies when the
user moves to another field or button, and ACTIVATE applies when the
user presses enter.

As a REALIZED callback:
Performs the same function as the above LOST_FOCUS and ACTIVATE
callbacks. This ensures that any initial value in the field is correctly
formatted. If there are any spaces in the field, for example, they are
removed. If the initial value of the field is not a proper integer
number, the MODIFY_VERIFY callback will cause trouble until the user
deletes the bad data, so attaching this as a realize callback is a
good idea.  Also, if ELECTROSTATIC is defined, the callback will
attach itself as a MODIFY_VERIFY, LOST_FOCUS, and ACTIVATE callback,
to save you some typing in PhAB.
-----------------------------------------------------------------------
*/

int verify_int( PtWidget_t *widget,
                ApInfo_t *apinfo,
                PtCallbackInfo_t *cbinfo )
{
    PtTextCallback_t *data;
    char              buff[120];
    char             *text;
    int               i;

    apinfo = apinfo;

    if ( cbinfo->reason == Pt_CB_MODIFY_VERIFY )
    {
        data = (PtTextCallback_t *)cbinfo->cbdata;

        data->doit = Pt_TRUE;

        make_str( widget, data, buff );

        if ( strlen( buff ) > 0 )
        {
            if ( buff[0] != '-' && !isdigit( buff[0] ) )
            {
                data->doit = Pt_FALSE;
            }

            for ( i=1; i < strlen( buff ) && data->doit; i++ )
            {
                if ( !isdigit( buff[i] ) )
                {
                    data->doit = Pt_FALSE;
                }
            }
        }
    }
    else
    if ( cbinfo->reason == Pt_CB_LOST_FOCUS ||
         cbinfo->reason == Pt_CB_ACTIVATE   ||
         cbinfo->reason == Pt_CB_REALIZED      )
    {
        text = get_text( widget, cbinfo );

        sprintf( buff, "%d", atoi( text ) );

        if ( strcmp( buff, text ) )
        {
            PtTextModifyText( widget, 0, strlen(text), 0, buff, strlen(buff) );
        }
    }

    #ifdef ELECTROSTATIC
        if ( cbinfo->reason == Pt_CB_REALIZED )
        {
            attach_callbacks( widget, &verify_int );
        }
    #endif

    return( Pt_CONTINUE );
}


/*
-----------------------------------------------------------------------
int verify_fint( widget, apinfo, cbinfo ) - verify & format integers.

Behaves like verify_int, except that a printf-style format string must
be attached to the widget as user data. The field can then be formatted
in this style whenever a REALIZED, ACTIVATE, or LOST_FOCUS callback
occurs. The format string must not cause the field to contain non-int
characters, or else the MODIFY_VERIFY callback will stop working
properly.
-----------------------------------------------------------------------
*/

int verify_fint( PtWidget_t *widget,
                 ApInfo_t *apinfo,
                 PtCallbackInfo_t *cbinfo )
{
    PtArg_t  arg;
    char     buff[120];
    char    *text;
    char    *format;

    if ( cbinfo->reason == Pt_CB_MODIFY_VERIFY )
    {
        verify_int( widget, apinfo, cbinfo );
    }
    else
    if ( cbinfo->reason == Pt_CB_LOST_FOCUS ||
         cbinfo->reason == Pt_CB_ACTIVATE   ||
         cbinfo->reason == Pt_CB_REALIZED      )
    {
        PtSetArg( &arg, Pt_ARG_USER_DATA, &format, 0 );
        PtGetResources( widget, 1, &arg );

        if ( format )
        {
            text = get_text( widget, cbinfo );

            sprintf( buff, format, atoi( text ) );

            if ( strcmp( buff, text ) )
            {
                PtTextModifyText( widget, 0, strlen(text),
                                  0, buff, strlen(buff) );
            }
        }
    }

    #ifdef ELECTROSTATIC
        if ( cbinfo->reason == Pt_CB_REALIZED )
        {
            attach_callbacks( widget, &verify_fint );
        }
    #endif

    return( Pt_CONTINUE );
}


/*
-----------------------------------------------------------------------
int verify_float( widget, apinfo, cbinfo ) - verify float values.

As a MODIFY_VERIFY callback:
Verifies that a text field contains a valid floating point value.  A
float can start with a '-' character and contain one decimal point,
but is otherwise all digits.

As a REALIZED, ACTIVATE, or LOST_FOCUS callback, this function behaves
similarly to verify_int().
-----------------------------------------------------------------------
*/

int verify_float( PtWidget_t *widget,
                  ApInfo_t *apinfo,
                  PtCallbackInfo_t *cbinfo )
{
    PtTextCallback_t *data;
    char              buff[120];
    char             *text;
    int               i, decimal_point;

    apinfo = apinfo;

    if ( cbinfo->reason == Pt_CB_MODIFY_VERIFY )
    {
        data = (PtTextCallback_t *)cbinfo->cbdata;

        data->doit = Pt_TRUE;

        make_str( widget, data, buff );

        if ( strlen( buff ) > 0 )
        {
            /* First char can be '-', '.', or a digit. */
            if ( buff[0] != '-' && buff[0] != '.' && !isdigit( buff[0] ) )
            {
                data->doit = Pt_FALSE;
            }

            /* Other chars can be '.' or digits */
            for ( i=1; i < strlen( buff ) && data->doit; i++ )
            {
                if ( buff[i] != '.' && !isdigit( buff[i] ) )
                {
                    data->doit = Pt_FALSE;
                }
            }

            /* There can be only 1 '.' char. */
            decimal_point = Pt_FALSE;
            for ( i=0; i < strlen( buff ) && data->doit; i++ )
            {
                if ( buff[i] == '.' )
                {
                    if ( decimal_point )
                    {
                        data->doit = Pt_FALSE;
                    }
                    else
                    {
                        decimal_point = Pt_TRUE;
                    }
                }
            }
        }
    }
    else
    if ( cbinfo->reason == Pt_CB_LOST_FOCUS ||
         cbinfo->reason == Pt_CB_ACTIVATE   ||
         cbinfo->reason == Pt_CB_REALIZED      )
    {
        text = get_text( widget, cbinfo );

        sprintf( buff, "%g", atof( text ) );

        if ( strcmp( buff, text ) )
        {
            PtTextModifyText( widget, 0, strlen(text), 0, buff, strlen(buff) );
        }
    }

    #ifdef ELECTROSTATIC
        if ( cbinfo->reason == Pt_CB_REALIZED )
        {
            attach_callbacks( widget, &verify_float );
        }
    #endif

    return( Pt_CONTINUE );
}


/*
-----------------------------------------------------------------------
int verify_ufloat( widget, apinfo, cbinfo ) - verify unsigned floats.

As a MODIFY_VERIFY callback:
Verifies that a text field contains a valid, positive, floating point
value.  A positive float can contain one decimal point, but is
otherwise all digits.

As a REALIZED, ACTIVATE, or LOST_FOCUS callback, this function behaves
similarly to verify_int().
-----------------------------------------------------------------------
*/

int verify_ufloat( PtWidget_t *widget,
                   ApInfo_t *apinfo,
                   PtCallbackInfo_t *cbinfo )
{
    PtTextCallback_t *data;
    char              buff[120];
    char             *text;
    int               i, decimal_point;

    apinfo = apinfo;

    if ( cbinfo->reason == Pt_CB_MODIFY_VERIFY )
    {
        data = (PtTextCallback_t *)cbinfo->cbdata;

        data->doit = Pt_TRUE;

        make_str( widget, data, buff );

        if ( strlen( buff ) > 0 )
        {
            /* Chars can be '.' or digits */
            for ( i=0; i < strlen( buff ) && data->doit; i++ )
            {
                if ( buff[i] != '.' && !isdigit( buff[i] ) )
                {
                    data->doit = Pt_FALSE;
                }
            }

            /* There can be only 1 '.' char. */
            decimal_point = Pt_FALSE;
            for ( i=0; i < strlen( buff ) && data->doit; i++ )
            {
                if ( buff[i] == '.' )
                {
                    if ( decimal_point )
                    {
                        data->doit = Pt_FALSE;
                    }
                    else
                    {
                        decimal_point = Pt_TRUE;
                    }
                }
            }
        }
    }
    if ( cbinfo->reason == Pt_CB_LOST_FOCUS ||
         cbinfo->reason == Pt_CB_ACTIVATE   ||
         cbinfo->reason == Pt_CB_REALIZED      )
    {
        text = get_text( widget, cbinfo );

        sprintf( buff, "%g", atof( text ) );

        if ( strcmp( buff, text ) )
        {
            PtTextModifyText( widget, 0, strlen(text), 0, buff, strlen(buff) );
        }
    }

    #ifdef ELECTROSTATIC
        if ( cbinfo->reason == Pt_CB_REALIZED )
        {
            attach_callbacks( widget, &verify_ufloat );
        }
    #endif

    return( Pt_CONTINUE );
}


/*
-----------------------------------------------------------------------
int verify_ffloat( widget, apinfo, cbinfo ) - verify & format floats.

Behaves like verify_fint(), but for floating point values. Remember
that a printf-style format string must be attached to the widget as
user data for this function to work.
-----------------------------------------------------------------------
*/

int verify_ffloat( PtWidget_t *widget,
                   ApInfo_t *apinfo,
                   PtCallbackInfo_t *cbinfo )
{
    PtArg_t  arg;
    char     buff[120];
    char    *text;
    char    *format;

    if ( cbinfo->reason == Pt_CB_MODIFY_VERIFY )
    {
        verify_float( widget, apinfo, cbinfo );
    }
    else
    if ( cbinfo->reason == Pt_CB_LOST_FOCUS ||
         cbinfo->reason == Pt_CB_ACTIVATE   ||
         cbinfo->reason == Pt_CB_REALIZED      )
    {
        PtSetArg( &arg, Pt_ARG_USER_DATA, &format, 0 );
        PtGetResources( widget, 1, &arg );

        if ( format )
        {
            text = get_text( widget, cbinfo );

            sprintf( buff, format, atof( text ) );

            if ( strcmp( buff, text ) )
            {
                PtTextModifyText( widget, 0, strlen(text),
                                  0, buff, strlen(buff) );
            }
        }
    }

    #ifdef ELECTROSTATIC
        if ( cbinfo->reason == Pt_CB_REALIZED )
        {
            attach_callbacks( widget, &verify_ffloat );
        }
    #endif

    return( Pt_CONTINUE );
}


/*
-----------------------------------------------------------------------
int verify_fufloat( widget, apinfo, cbinfo ) - verify & format ufloats.

Behaves like verify_ffloat(), but negative numbers are not allowed.
Remember that a printf-style format string must be attached to the
widget as user data for this function to work.
-----------------------------------------------------------------------
*/

int verify_fufloat( PtWidget_t *widget,
                    ApInfo_t *apinfo,
                    PtCallbackInfo_t *cbinfo )
{
    PtArg_t  arg;
    char     buff[120];
    char    *text;
    char    *format;

    if ( cbinfo->reason == Pt_CB_MODIFY_VERIFY )
    {
        verify_ufloat( widget, apinfo, cbinfo );
    }
    else
    if ( cbinfo->reason == Pt_CB_LOST_FOCUS ||
         cbinfo->reason == Pt_CB_ACTIVATE   ||
         cbinfo->reason == Pt_CB_REALIZED      )
    {
        PtSetArg( &arg, Pt_ARG_USER_DATA, &format, 0 );
        PtGetResources( widget, 1, &arg );

        if ( format )
        {
            text = get_text( widget, cbinfo );

            sprintf( buff, format, fabs( atof( text ) ) );

            if ( strcmp( buff, text ) )
            {
                PtTextModifyText( widget, 0, strlen(text),
                                  0, buff, strlen(buff) );
            }
        }
    }

    #ifdef ELECTROSTATIC
        if ( cbinfo->reason == Pt_CB_REALIZED )
        {
            attach_callbacks( widget, &verify_fufloat );
        }
    #endif

    return( Pt_CONTINUE );
}


/*
-----------------------------------------------------------------------
int verify_unsigned( widget, apinfo, cbinfo ) - verify unsigned ints.

As a MODIFY_VERIFY callback:
Verifies that a text field contains a valid unsigned integer value.
An unsigned integer can contain only digits.

As a REALIZED, ACTIVATE or LOST_FOCUS callback:
Behaves like verify_int().
-----------------------------------------------------------------------
*/

int verify_unsigned( PtWidget_t *widget,
                     ApInfo_t *apinfo,
                     PtCallbackInfo_t *cbinfo )
{
    PtTextCallback_t *data;
    char              buff[120];
    char             *text;
    int               i;

    apinfo = apinfo;

    if ( cbinfo->reason == Pt_CB_MODIFY_VERIFY )
    {
        data = (PtTextCallback_t *)cbinfo->cbdata;

        data->doit = Pt_TRUE;

        make_str( widget, data, buff );

        for ( i=0; i < strlen( buff ) && data->doit; i++ )
        {
            if ( !isdigit( buff[i] ) )
            {
                data->doit = Pt_FALSE;
            }
        }
    }
    else
    if ( cbinfo->reason == Pt_CB_LOST_FOCUS ||
         cbinfo->reason == Pt_CB_ACTIVATE   ||
         cbinfo->reason == Pt_CB_REALIZED      )
    {
        text = get_text( widget, cbinfo );

        sprintf( buff, "%u", max( 0, atoi( text ) ) );

        if ( strcmp( buff, text ) )
        {
            PtTextModifyText( widget, 0, strlen(text), 0, buff, strlen(buff) );
        }
    }

    #ifdef ELECTROSTATIC
        if ( cbinfo->reason == Pt_CB_REALIZED )
        {
            attach_callbacks( widget, &verify_unsigned );
        }
    #endif

    return( Pt_CONTINUE );
}


/*
-----------------------------------------------------------------------
int verify_funsigned( widget, apinfo, cbinfo ) - verify & format uints.

Behaves like verify_fint(), but for unsigned integer values. Remember
that a printf-style format string must be attached to the widget as
user data for this function to work.
-----------------------------------------------------------------------
*/

int verify_funsigned( PtWidget_t *widget,
                      ApInfo_t *apinfo,
                      PtCallbackInfo_t *cbinfo )
{
    PtArg_t  arg;
    char     buff[120];
    char    *text;
    char    *format;

    if ( cbinfo->reason == Pt_CB_MODIFY_VERIFY )
    {
        verify_unsigned( widget, apinfo, cbinfo );
    }
    else
    if ( cbinfo->reason == Pt_CB_LOST_FOCUS ||
         cbinfo->reason == Pt_CB_ACTIVATE   ||
         cbinfo->reason == Pt_CB_REALIZED      )
    {
        PtSetArg( &arg, Pt_ARG_USER_DATA, &format, 0 );
        PtGetResources( widget, 1, &arg );

        if ( format )
        {
            text = get_text( widget, cbinfo );

            sprintf( buff, format, max( 0, atoi( text ) ) );

            if ( strcmp( buff, text ) )
            {
                PtTextModifyText( widget, 0, strlen(text),
                                  0, buff, strlen(buff) );
            }
        }
    }

    #ifdef ELECTROSTATIC
        if ( cbinfo->reason == Pt_CB_REALIZED )
        {
            attach_callbacks( widget, &verify_funsigned );
        }
    #endif

    return( Pt_CONTINUE );
}


/*
-----------------------------------------------------------------------
int verify_unumhex( widget, apinfo, cbinfo ) - verify dec/hex unsigned.

As a MODIFY_VERIFY callback:
Verifies that a text field contains a valid unsigned integer value, in
decimal or hex notation.
If the number begins with 0x, digits 0-f or 0-F are allowed.
Otherwise only digits 0-9 are allowed.

As a REALIZED, ACTIVATE or LOST_FOCUS callback:
Ensures the field contains a valid unsigned hex (0x[0-f]*) or decimal
([0-9]*) number.
Also, performs the same callback-attaching functions as verify_int().
-----------------------------------------------------------------------
*/

int verify_unumhex( PtWidget_t *widget,
                    ApInfo_t *apinfo,
                    PtCallbackInfo_t *cbinfo )
{
    PtTextCallback_t *data;
    char              buff[120];
    char             *text;
    int               hex;
    int               i;

    apinfo = apinfo;

    if ( cbinfo->reason == Pt_CB_MODIFY_VERIFY )
    {
        data = (PtTextCallback_t *)cbinfo->cbdata;

        data->doit = Pt_TRUE;

        make_str( widget, data, buff );

        if ( !strnicmp( "0x", buff, strlen("0x") ) )
            hex = Pt_TRUE;
        else
            hex = Pt_FALSE;

        if ( hex )
        {
            for ( i=strlen("0x"); i < strlen( buff ) && data->doit; i++ )
            {
                if ( !isxdigit( buff[i] ) )
                {
                    data->doit = Pt_FALSE;
                }
            }
        }
        else
        {
            for ( i=0; i < strlen( buff ) && data->doit; i++ )
            {
                if ( !isdigit( buff[i] ) )
                {
                    data->doit = Pt_FALSE;
                }
            }
        }
    }
    else
    if ( cbinfo->reason == Pt_CB_LOST_FOCUS ||
         cbinfo->reason == Pt_CB_ACTIVATE   ||
         cbinfo->reason == Pt_CB_REALIZED      )
    {
        text = get_text( widget, cbinfo );

        if ( !strnicmp( "0x", text, strlen("0x") ) )
            hex = Pt_TRUE;
        else
            hex = Pt_FALSE;

        if ( hex )
        {
            sprintf( buff, "0x%x", strtol( text, NULL, 0 ) );
        }
        else
        {
            sprintf( buff, "%u", max( 0, atoi( text ) ) );
        }

        if ( stricmp( buff, text ) )
        {
            PtTextModifyText( widget, 0, strlen(text), 0, buff, strlen(buff) );
        }
    }

    #ifdef ELECTROSTATIC
        if ( cbinfo->reason == Pt_CB_REALIZED )
        {
            attach_callbacks( widget, &verify_unumhex );
        }
    #endif

    return( Pt_CONTINUE );
}


/*
-----------------------------------------------------------------------
int verify_alpha( widget, apinfo, cbinfo ) - verify alphabetic strings.

As a MODIFY_VERIFY callback:
Allows only alphabetic characters into the text widget. Uses isalpha()
to test input.

As a LOST_FOCUS or ACTIVATE callback:
Does nothing.

As a REALIZED callback:
Strips out any non-alpha characters in the initial value of the field.
Also, if ELECTROSTATIC is defined, the callback will attach itself as
a MODIFY_VERIFY, LOST_FOCUS, and ACTIVATE callback, like verify_int()
and the other numeric verify_* functions.
-----------------------------------------------------------------------
*/

int verify_alpha( PtWidget_t *widget,
                  ApInfo_t *apinfo,
                  PtCallbackInfo_t *cbinfo )
{
    PtTextCallback_t *data;
    char              buff[120];
    char             *text;
    char             *p, *q;
    int               i;

    apinfo = apinfo;

    if ( cbinfo->reason == Pt_CB_MODIFY_VERIFY )
    {
        data = (PtTextCallback_t *)cbinfo->cbdata;

        data->doit = Pt_TRUE;

        make_str( widget, data, buff );

        for ( i=0; i < strlen( buff ) && data->doit; i++ )
        {
            if ( !isalpha( buff[i] ) )
            {
                data->doit = Pt_FALSE;
            }
        }
    }
    else
    if ( cbinfo->reason == Pt_CB_REALIZED )
    {
        text = get_text( widget, cbinfo );

        for ( p=text, q=&buff, i=0; *p != NULL && i < sizeof(buff)-1; p++ )
        {
            if ( isalpha(*p) )
            {
                *q = *p;
                q++;
                i++;
            }
        }
        *q = NULL;

        if ( strcmp( buff, text ) )
        {
            PtTextModifyText( widget, 0, strlen(text), 0, buff, strlen(buff) );
        }
    }

    #ifdef ELECTROSTATIC
        if ( cbinfo->reason == Pt_CB_REALIZED )
        {
            attach_callbacks( widget, &verify_alpha );
        }
    #endif

    return( Pt_CONTINUE );
}


/*
-----------------------------------------------------------------------
int verify_alphanum( widget, apinfo, cbinfo ) - verify alphanumeric strings.

As a MODIFY_VERIFY callback:
Allows only alphanumeric characters into the text widget. Uses isalnum()
to test input.

As a REALIZED, LOST_FOCUS or ACTIVATE callback:
Behaves like verify_alpha().
-----------------------------------------------------------------------
*/

int verify_alphanum( PtWidget_t *widget,
                     ApInfo_t *apinfo,
                     PtCallbackInfo_t *cbinfo )
{
    PtTextCallback_t *data;
    char              buff[120];
    char             *text;
    char             *p, *q;
    int               i;

    apinfo = apinfo;

    if ( cbinfo->reason == Pt_CB_MODIFY_VERIFY )
    {
        data = (PtTextCallback_t *)cbinfo->cbdata;

        data->doit = Pt_TRUE;

        make_str( widget, data, buff );

        for ( i=0; i < strlen( buff ) && data->doit; i++ )
        {
            if ( !isalnum( buff[i] ) )
            {
                data->doit = Pt_FALSE;
            }
        }
    }
    else
    if ( cbinfo->reason == Pt_CB_REALIZED )
    {
        text = get_text( widget, cbinfo );

        for ( p=text, q=&buff, i=0; *p != NULL && i < sizeof(buff)-1; p++ )
        {
            if ( isalnum(*p) )
            {
                *q = *p;
                q++;
                i++;
            }
        }
        *q = NULL;

        if ( strcmp( buff, text ) )
        {
            PtTextModifyText( widget, 0, strlen(text), 0, buff, strlen(buff) );
        }
    }

    #ifdef ELECTROSTATIC
        if ( cbinfo->reason == Pt_CB_REALIZED )
        {
            attach_callbacks( widget, &verify_alphanum );
        }
    #endif

    return( Pt_CONTINUE );
}


/*
-----------------------------------------------------------------------
int verify_upper( widget, apinfo, cbinfo ) - verify upper case strings.

As a MODIFY_VERIFY callback:
Allows only upper case alphabetic characters into the text widget. Uses
isalpha() to test input, and converts any lower case entries into upper
case.

As a REALIZED, LOST_FOCUS or ACTIVATE callback:
Behaves like verify_alpha().
-----------------------------------------------------------------------
*/

int verify_upper( PtWidget_t *widget,
                  ApInfo_t *apinfo,
                  PtCallbackInfo_t *cbinfo )
{
    PtTextCallback_t *data;
    char              buff[120];
    char             *text;
    char             *p, *q;
    int               i;

    apinfo = apinfo;

    if ( cbinfo->reason == Pt_CB_MODIFY_VERIFY )
    {
        data = (PtTextCallback_t *)cbinfo->cbdata;

        data->doit = Pt_TRUE;

        make_str( widget, data, buff );

        for ( i=0; i < strlen( buff ) && data->doit; i++ )
        {
            if ( !isalpha( buff[i] ) )
            {
                data->doit = Pt_FALSE;
            }
        }

        for ( i=0; i < data->length; i++ )
        {
            data->text[i] = toupper( data->text[i] );
        }
    }
    else
    if ( cbinfo->reason == Pt_CB_REALIZED )
    {
        text = get_text( widget, cbinfo );

        for ( p=text, q=&buff, i=0; *p != NULL && i < sizeof(buff)-1; p++ )
        {
            if ( isalpha(*p) )
            {
                *q = toupper( *p );
                q++;
                i++;
            }
        }
        *q = NULL;

        if ( strcmp( buff, text ) )
        {
            PtTextModifyText( widget, 0, strlen(text), 0, buff, strlen(buff) );
        }
    }

    #ifdef ELECTROSTATIC
        if ( cbinfo->reason == Pt_CB_REALIZED )
        {
            attach_callbacks( widget, &verify_upper );
        }
    #endif

    return( Pt_CONTINUE );
}


/*
-----------------------------------------------------------------------
int verify_lower( widget, apinfo, cbinfo ) - verify lower case strings.

As a MODIFY_VERIFY callback:
Allows only lower case alphabetic characters into the text widget. Uses
isalpha() to test input, and converts any upper case entries into lower
case.

As a REALIZED, LOST_FOCUS or ACTIVATE callback:
Behaves like verify_alpha().
-----------------------------------------------------------------------
*/

int verify_lower( PtWidget_t *widget,
                  ApInfo_t *apinfo,
                  PtCallbackInfo_t *cbinfo )
{
    PtTextCallback_t *data;
    char              buff[120];
    char             *text;
    char             *p, *q;
    int               i;

    apinfo = apinfo;

    if ( cbinfo->reason == Pt_CB_MODIFY_VERIFY )
    {
        data = (PtTextCallback_t *)cbinfo->cbdata;

        data->doit = Pt_TRUE;

        make_str( widget, data, buff );

        for ( i=0; i < strlen( buff ) && data->doit; i++ )
        {
            if ( !isalpha( buff[i] ) )
            {
                data->doit = Pt_FALSE;
            }
        }

        for ( i=0; i < data->length; i++ )
        {
            data->text[i] = tolower( data->text[i] );
        }
    }
    else
    if ( cbinfo->reason == Pt_CB_REALIZED )
    {
        text = get_text( widget, cbinfo );

        for ( p=text, q=&buff, i=0; *p != NULL && i < sizeof(buff)-1; p++ )
        {
            if ( isalpha(*p) )
            {
                *q = tolower( *p );
                q++;
                i++;
            }
        }
        *q = NULL;

        if ( strcmp( buff, text ) )
        {
            PtTextModifyText( widget, 0, strlen(text), 0, buff, strlen(buff) );
        }
    }

    #ifdef ELECTROSTATIC
        if ( cbinfo->reason == Pt_CB_REALIZED )
        {
            attach_callbacks( widget, &verify_lower );
        }
    #endif

    return( Pt_CONTINUE );
}


/*
-----------------------------------------------------------------------
int verify_uppernum( widget, apinfo, cbinfo ) - upper case & numbers.

As a MODIFY_VERIFY callback:
Allows only upper case alphabetic characters into the text widget. Uses
isalnum() to test input, and converts any lower case entries into upper
case.

As a REALIZED, LOST_FOCUS or ACTIVATE callback:
Behaves like verify_alpha().
-----------------------------------------------------------------------
*/

int verify_uppernum( PtWidget_t *widget,
                     ApInfo_t *apinfo,
                     PtCallbackInfo_t *cbinfo )
{
    PtTextCallback_t *data;
    char              buff[120];
    char             *text;
    char             *p, *q;
    int               i;

    apinfo = apinfo;

    if ( cbinfo->reason == Pt_CB_MODIFY_VERIFY )
    {
        data = (PtTextCallback_t *)cbinfo->cbdata;

        data->doit = Pt_TRUE;

        make_str( widget, data, buff );

        for ( i=0; i < strlen( buff ) && data->doit; i++ )
        {
            if ( !isalnum( buff[i] ) )
            {
                data->doit = Pt_FALSE;
            }
        }

        for ( i=0; i < data->length; i++ )
        {
            data->text[i] = toupper( data->text[i] );
        }
    }
    else
    if ( cbinfo->reason == Pt_CB_REALIZED )
    {
        text = get_text( widget, cbinfo );

        for ( p=text, q=&buff, i=0; *p != NULL && i < sizeof(buff)-1; p++ )
        {
            if ( isalnum(*p) )
            {
                *q = toupper( *p );
                q++;
                i++;
            }
        }
        *q = NULL;

        if ( strcmp( buff, text ) )
        {
            PtTextModifyText( widget, 0, strlen(text), 0, buff, strlen(buff) );
        }
    }

    #ifdef ELECTROSTATIC
        if ( cbinfo->reason == Pt_CB_REALIZED )
        {
            attach_callbacks( widget, &verify_uppernum );
        }
    #endif

    return( Pt_CONTINUE );
}


/*
-----------------------------------------------------------------------
int verify_lowernum( widget, apinfo, cbinfo ) - lower case & numbers.

As a MODIFY_VERIFY callback:
Allows only lower case alphabetic characters into the text widget. Uses
isalnum() to test input, and converts any upper case entries into lower
case.

As a REALIZED, LOST_FOCUS or ACTIVATE callback:
Behaves like verify_alpha().
-----------------------------------------------------------------------
*/

int verify_lowernum( PtWidget_t *widget,
                     ApInfo_t *apinfo,
                     PtCallbackInfo_t *cbinfo )
{
    PtTextCallback_t *data;
    char              buff[120];
    char             *text;
    char             *p, *q;
    int               i;

    apinfo = apinfo;

    if ( cbinfo->reason == Pt_CB_MODIFY_VERIFY )
    {
        data = (PtTextCallback_t *)cbinfo->cbdata;

        data->doit = Pt_TRUE;

        make_str( widget, data, buff );

        for ( i=0; i < strlen( buff ) && data->doit; i++ )
        {
            if ( !isalnum( buff[i] ) )
            {
                data->doit = Pt_FALSE;
            }
        }

        for ( i=0; i < data->length; i++ )
        {
            data->text[i] = tolower( data->text[i] );
        }
    }
    else
    if ( cbinfo->reason == Pt_CB_REALIZED )
    {
        text = get_text( widget, cbinfo );

        for ( p=text, q=&buff, i=0; *p != NULL && i < sizeof(buff)-1; p++ )
        {
            if ( isalnum(*p) )
            {
                *q = tolower( *p );
                q++;
                i++;
            }
        }
        *q = NULL;

        if ( strcmp( buff, text ) )
        {
            PtTextModifyText( widget, 0, strlen(text), 0, buff, strlen(buff) );
        }
    }

    #ifdef ELECTROSTATIC
        if ( cbinfo->reason == Pt_CB_REALIZED )
        {
            attach_callbacks( widget, &verify_lowernum );
        }
    #endif

    return( Pt_CONTINUE );
}


/*
-----------------------------------------------------------------------
int verify_file( widget, apinfo, cbinfo ) - verify filename safe strings.

As a MODIFY_VERIFY callback:
Allows only alphanumerics and '.' and '_' characters into the text
widget.

As a REALIZED, LOST_FOCUS or ACTIVATE callback:
Behaves like verify_alpha().
-----------------------------------------------------------------------
*/

int verify_file( PtWidget_t *widget,
                 ApInfo_t *apinfo,
                 PtCallbackInfo_t *cbinfo )
{
    PtTextCallback_t *data;
    char              buff[120];
    char             *text;
    char             *p, *q;
    int               i;

    apinfo = apinfo;

    if ( cbinfo->reason == Pt_CB_MODIFY_VERIFY )
    {
        data = (PtTextCallback_t *)cbinfo->cbdata;

        data->doit = Pt_TRUE;

        make_str( widget, data, buff );

        for ( i=0; i < strlen( buff ) && data->doit; i++ )
        {
            if ( !isalnum( buff[i] ) && !strchr( "._", buff[i] ) )
            {
                data->doit = Pt_FALSE;
            }
        }
    }
    else
    if ( cbinfo->reason == Pt_CB_REALIZED )
    {
        text = get_text( widget, cbinfo );

        for ( p=text, q=&buff, i=0; *p != NULL && i < sizeof(buff)-1; p++ )
        {
            if ( isalnum(*p) || strchr( "._", text[i] ) )
            {
                *q = *p;
                q++;
                i++;
            }
        }
        *q = NULL;

        if ( strcmp( buff, text ) )
        {
            PtTextModifyText( widget, 0, strlen(text), 0, buff, strlen(buff) );
        }
    }

    #ifdef ELECTROSTATIC
        if ( cbinfo->reason == Pt_CB_REALIZED )
        {
            attach_callbacks( widget, &verify_file );
        }
    #endif

    return( Pt_CONTINUE );
}


/*
-----------------------------------------------------------------------
int verify_path( widget, apinfo, cbinfo ) - verify pathname safe strings.

As a MODIFY_VERIFY callback:
Allows only alphanumerics, '/', '.' and '_' characters into the text
widget. Does not currently test for sensible pathnames; '/////' is
accepted.

As a REALIZED, LOST_FOCUS or ACTIVATE callback:
Behaves like verify_alpha().
-----------------------------------------------------------------------
*/

int verify_path( PtWidget_t *widget,
                 ApInfo_t *apinfo,
                 PtCallbackInfo_t *cbinfo )
{
    PtTextCallback_t *data;
    char              buff[120];
    char             *text;
    char             *p, *q;
    int               i;

    apinfo = apinfo;

    if ( cbinfo->reason == Pt_CB_MODIFY_VERIFY )
    {
        data = (PtTextCallback_t *)cbinfo->cbdata;

        data->doit = Pt_TRUE;

        make_str( widget, data, buff );

        for ( i=0; i < strlen( buff ) && data->doit; i++ )
        {
            if ( !isalnum( buff[i] ) && !strchr( "._/", buff[i] ) )
            {
                data->doit = Pt_FALSE;
            }
        }
    }
    else
    if ( cbinfo->reason == Pt_CB_REALIZED )
    {
        text = get_text( widget, cbinfo );

        for ( p=text, q=&buff, i=0; *p != NULL && i < sizeof(buff)-1; p++ )
        {
            if ( isalnum(*p) || strchr( "._/", buff[i] ) )
            {
                *q = *p;
                q++;
                i++;
            }
        }
        *q = NULL;

        if ( strcmp( buff, text ) )
        {
            PtTextModifyText( widget, 0, strlen(text), 0, buff, strlen(buff) );
        }
    }

    #ifdef ELECTROSTATIC
        if ( cbinfo->reason == Pt_CB_REALIZED )
        {
            attach_callbacks( widget, &verify_path );
        }
    #endif

    return( Pt_CONTINUE );
}


/*
-----------------------------------------------------------------------
int verify_charset( widget, apinfo, cbinfo ) - verify against a charset.

As a MODIFY_VERIFY callback:
Verifies that data entry contains only characters from a specified
character set. The character set must be attached to the widget as 
user data. It is specified as a NULL-terminated string of acceptable
characters. The callback is case-sensitive, so mixed-case text entry
is only possible if both upper and lower case characters are present
in the charset.

As a REALIZED, LOST_FOCUS or ACTIVATE callback:
Behaves like verify_alpha().
-----------------------------------------------------------------------
*/

int verify_charset( PtWidget_t *widget,
                    ApInfo_t *apinfo,
                    PtCallbackInfo_t *cbinfo )
{
    PtTextCallback_t *data;
    PtArg_t           arg;
    char             *charset;
    char              buff[120];
    char             *text;
    char             *p, *q;
    int               i;

    apinfo = apinfo;

    PtSetArg( &arg, Pt_ARG_USER_DATA, &charset, 0 );
    PtGetResources( widget, 1, &arg );

    if ( !charset )
    {
        return ( Pt_CONTINUE );
    }

    if ( cbinfo->reason == Pt_CB_MODIFY_VERIFY )
    {
        data = (PtTextCallback_t *)cbinfo->cbdata;

        data->doit = Pt_TRUE;

        make_str( widget, data, buff );

        for ( i=0; i < strlen( buff ) && data->doit; i++ )
        {
            if ( !strchr( charset, buff[i] ) )
            {
                data->doit = Pt_FALSE;
            }
        }
    }
    else
    if ( cbinfo->reason == Pt_CB_REALIZED )
    {
        text = get_text( widget, cbinfo );

        for ( p=text, q=&buff, i=0; *p != NULL && i < sizeof(buff)-1; p++ )
        {
            if ( strchr( charset, *p) )
            {
                *q = *p;
                q++;
                i++;
            }
        }
        *q = NULL;

        if ( strcmp( buff, text ) )
        {
            PtTextModifyText( widget, 0, strlen(text), 0, buff, strlen(buff) );
        }
    }

    #ifdef ELECTROSTATIC
        if ( cbinfo->reason == Pt_CB_REALIZED )
        {
            attach_callbacks( widget, &verify_charset );
        }
    #endif

    return( Pt_CONTINUE );
}


/*
-----------------------------------------------------------------------
int verify_upperset( widget, apinfo, cbinfo ) - upper case charset.

As a MODIFY_VERIFY callback:
Verifies that data entry contains only characters from a specified
character set. The character set must be attached to the widget as 
user data. It is specified as a NULL-terminated string of acceptable
characters. Lower case characters in the charset are ignored, and
lower case characters in the input data are converted to upper case.

As a REALIZED, LOST_FOCUS or ACTIVATE callback:
Behaves like verify_alpha().
-----------------------------------------------------------------------
*/

int verify_upperset( PtWidget_t *widget,
                     ApInfo_t *apinfo,
                     PtCallbackInfo_t *cbinfo )
{
    PtTextCallback_t *data;
    PtArg_t           arg;
    char             *charset;
    char              buff[120];
    char             *text;
    char             *p, *q;
    int               i;

    apinfo = apinfo;

    PtSetArg( &arg, Pt_ARG_USER_DATA, &charset, 0 );
    PtGetResources( widget, 1, &arg );

    if ( !charset )
    {
        return ( Pt_CONTINUE );
    }

    if ( cbinfo->reason == Pt_CB_MODIFY_VERIFY )
    {
        data = (PtTextCallback_t *)cbinfo->cbdata;

        data->doit = Pt_TRUE;

        make_str( widget, data, buff );

        for ( i=0; i < strlen( buff ) && data->doit; i++ )
        {
            if ( !strchr( charset, toupper(buff[i]) ) )
            {
                data->doit = Pt_FALSE;
            }
        }

        for ( i=0; i < data->length; i++ )
        {
            data->text[i] = toupper( data->text[i] );
        }
    }
    else
    if ( cbinfo->reason == Pt_CB_REALIZED )
    {
        text = get_text( widget, cbinfo );

        for ( p=text, q=&buff, i=0; *p != NULL && i < sizeof(buff)-1; p++ )
        {
            if ( strchr( charset, *p ) )
            {
                *q = toupper( *p );
                q++;
                i++;
            }
        }
        *q = NULL;

        if ( strcmp( buff, text ) )
        {
            PtTextModifyText( widget, 0, strlen(text), 0, buff, strlen(buff) );
        }
    }

    #ifdef ELECTROSTATIC
        if ( cbinfo->reason == Pt_CB_REALIZED )
        {
            attach_callbacks( widget, &verify_upperset );
        }
    #endif

    return( Pt_CONTINUE );
}


/*
-----------------------------------------------------------------------
int verify_lowerset( widget, apinfo, cbinfo ) - lower case charset.

As a MODIFY_VERIFY callback:
Verifies that data entry contains only characters from a specified
character set. The character set must be attached to the widget as 
user data. It is specified as a NULL-terminated string of acceptable
characters. Upper case characters in the charset are ignored, and
upper case characters in the input data are converted to lower case.

As a REALIZED, LOST_FOCUS or ACTIVATE callback:
Behaves like verify_alpha().
-----------------------------------------------------------------------
*/

int verify_lowerset( PtWidget_t *widget,
                     ApInfo_t *apinfo,
                     PtCallbackInfo_t *cbinfo )
{
    PtTextCallback_t *data;
    PtArg_t           arg;
    char             *charset;
    char              buff[120];
    char             *text;
    char             *p, *q;
    int               i;

    apinfo = apinfo;

    PtSetArg( &arg, Pt_ARG_USER_DATA, &charset, 0 );
    PtGetResources( widget, 1, &arg );

    if ( !charset )
    {
        return ( Pt_CONTINUE );
    }

    if ( cbinfo->reason == Pt_CB_MODIFY_VERIFY )
    {
        data = (PtTextCallback_t *)cbinfo->cbdata;

        data->doit = Pt_TRUE;

        make_str( widget, data, buff );

        for ( i=0; i < strlen( buff ) && data->doit; i++ )
        {
            if ( !strchr( charset, tolower(buff[i]) ) )
            {
                data->doit = Pt_FALSE;
            }
        }

        for ( i=0; i < data->length; i++ )
        {
            data->text[i] = tolower( data->text[i] );
        }
    }
    else
    if ( cbinfo->reason == Pt_CB_REALIZED )
    {
        text = get_text( widget, cbinfo );

        for ( p=text, q=&buff, i=0; *p != NULL && i < sizeof(buff)-1; p++ )
        {
            if ( strchr( charset, *p ) )
            {
                *q = tolower( *p );
                q++;
                i++;
            }
        }
        *q = NULL;

        if ( strcmp( buff, text ) )
        {
            PtTextModifyText( widget, 0, strlen(text), 0, buff, strlen(buff) );
        }
    }

    #ifdef ELECTROSTATIC
        if ( cbinfo->reason == Pt_CB_REALIZED )
        {
            attach_callbacks( widget, &verify_lowerset );
        }
    #endif

    return( Pt_CONTINUE );
}


